export default class AST_RENDERER {

    constructor(oGlobalTemplates, oGlobalContext, oLocalTemplates, oLocalContext) {

        this.mTemplates = new Map();

        // Will be set when constructed
        this.oData = null;
        this.vTempData = null;

        this.registerTemplates(oGlobalTemplates);

        if (oLocalTemplates) {
            this.registerTemplates(oLocalTemplates, true);
        }

        console.log(this.mTemplates);
    }

    registerTemplates(oTemplates, bLocal = false) {

        if (oTemplates) {

            for (let sTemplate in oTemplates) {

                if (!this.mTemplates.has(sTemplate)) {

                    this.mTemplates.set(sTemplate, oTemplates[sTemplate]);
                }
                else {

                    console.error(`Attempted to recreate a template: ${sTemplate}`);
                }

            }

        }

    }

    convertObjectToString(vObject) {

        let sResolvedValue = null;

        if (Array.isArray(vObject)) {

            sResolvedValue = "[Array]";
        }
        else if (vObject === null) {

            sResolvedValue = "[Null]";
        }
        else {

            sResolvedValue = "[Object]";
        }

        return sResolvedValue;
    }
  
    resolveLogic(oLogic, oData) {

        console.log(oLogic);

        const F_EXTRACT_EXP_VALUE = (oExpression, oData) => {

            // First evaluated eval
            switch (oExpression.eval.type) {

                case "literal":
                    
                    return oExpression.eval.value;

                case "data":

                    return this.resolveData(oExpression.eval, oData, false);

                default:

                    console.error("Unknown expression type: ${oExpression.eval.value}");

            }


        }

        const F_RESOLVE_CONDITION = (oCondition, oData) => {

            if (oCondition.type === "unary") {
    
                let vExpressValue = F_EXTRACT_EXP_VALUE(oCondition.expression, oData);

                if (oCondition.expression.inverse) {

                    return (vExpressValue) ? false: true;
                }
                else {

                    return (vExpressValue) ? true : false;
                }

            }
            else if (oCondition.type === "binary") {

                let vLeftEval = F_EXTRACT_EXP_VALUE(oCondition.expression.left, oData);
                let vRightEval = F_EXTRACT_EXP_VALUE(oCondition.expression.right, oData);

                // We will assume the result is false and toggle to true if needed!
                let bResult = false;

                switch (oCondition.expression.op) {

                    case "==":

                        if (vLeftEval == vRightEval) {
                            bResult = true;
                        }
                        break;

                    case "===":

                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }
                        break;

                    case "!=":
                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }  
                        break;

                    case "!==":
                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }  
                        break;

                    case ">":

                        if (vLeftEval > vRightEval) {
                            bResult = true;
                        } 
                        break;
                        
                    case ">=":

                        if (vLeftEval >= vRightEval) {
                            bResult = true;
                        } 
                        break;

                    case "<":

                        if (vLeftEval < vRightEval) {
                            bResult = true;
                        } 
                        break;


                    case "<=":

                        if (vLeftEval <= vRightEval) {
                            bResult = true;
                        } 
                        break;

                }

                return bResult;
            }

        }

        // Loop through all of the supplied conditionals, as 
        for (let oConditional of oLogic.conditionals) {

            if (F_RESOLVE_CONDITION(oConditional.conditional, oData)) {
                return oConditional.consequence
            }

        }

        if (oLogic.alternative && oLogic.alternative.length) {

            if (oLogic.alternative.length === 1 && Object.keys(oLogic.alternative).length === 1 && (oLogic.alternative[0].contents && oLogic.alternative[0].contents.length)) {

                return oLogic.alternative[0].contents;
            }


            return oLogic.alternative;
        }

        return null;
    }

    resolveData(oNode, oData, bError = true) {

        console.log(oNode);
        console.log(oData);

        // Continue only if the contents node for data has a string with length
        if (oNode && typeof oNode.value === "string" && oNode.value.length) {

            let oContext = {};

            if (this.oGlobalContext) {
                oContext = {
                    ...this.oGlobalContext
                };
            }

            if (typeof oData === "object") {
                oContext = {
                    ...oContext,
                    ...oData
                };
            }

            let asPath = oNode.value.split('.');

            if (asPath.length > 1 && asPath[0] === "this") {

                asPath = asPath.slice(1);
            }

            // Loop through the paths
            for (let p = 0, pLen = asPath.length; p < pLen; p++) {

                // Check to see if this part of the path exists!
                if (oContext.hasOwnProperty(asPath[p])) {

                    if (p + 1 === pLen) {

                        console.log(typeof oContext[asPath[p]]);
                        console.log(oContext[asPath[p]]);

                        return oContext[asPath[p]];
                    }
                    else {

                        oContext = oContext[asPath[p]];
                        continue;
                    }

                }
                else {

                    if (bError) {
                        console.error(`Requested invalid data context path: ${oNode.value}`);
                    }

                    return null;
                }

            }

        }

    }

    generateAttributes(aoAttributes, oData) {

        let oAttributes = {};

        const F_HANDLE_LITERAL_ATTR = (oLiteralNode, sAttrName) => {

            switch (typeof oLiteralNode.value) {

                case "string":
                    oAttributes[sAttrName].push(oLiteralNode.value);
                    break;

                case "number":
                case "boolean":
                    oAttributes[osAttrName].push(oLiteralNode.value.toString());
                    break;
            }

        }

        const F_HANDLE_DATA_ATTR = (oDataNode, sAttrName, oData) => {

            let vDataValue = this.resolveData(oDataNode, oData);

            switch(typeof vDataValue) {

                case "boolean":
                case "number":
                    vDataValue = vDataValue.toString();
                    break;
                
                case "object":

                    vDataValue = this.convertObjectToString(vDataValue);
                    break;

            }

            if (vDataValue) {

                oAttributes[sAttrName].push(vDataValue);
            }

        }

        const F_LOOP_ATTRIBUTES = (aoAttributes, sSpecificAttr = null, oData) => {

            // Loop through attribute objects
            for (let oNode of aoAttributes) {
                
                // Shift through all the object types
                switch (oNode.type) {
        
                    // Handle static typed attributes, or those that have dynamics inside of the attribute
                    case "attr":

                        let sAttributeName = null;

                        if (!sAttributeName) {
                            
                            sAttributeName = oNode.name;
                        }
                        else {

                            sAttributeName = sSpecificAttr;
                        }
                    
                        if (!oAttributes[sAttributeName]) {
                            oAttributes[sAttributeName] = [];
                        }
    
                        if (oNode.contents && oNode.contents.length) {
    
                            for (let av = 0, avLen = oNode.contents.length; av < avLen; av++) {
    
                                switch (oNode.contents[av].type) {
    
                                    case "literal":
    
                                        F_HANDLE_LITERAL_ATTR(oNode.contents[av], sAttributeName);
                                        break
    
                                    case "data":

                                        
                                        F_HANDLE_DATA_ATTR(oNode.contents[av], sAttributeName, oData);
                                        break;
                                        
                                    case "logic":
                                    case "logical":

                                        // console.log("Attribute logic");
                                        // console.log(oNode);
    
                                        let aoLogicResolve = this.resolveLogic(oNode.contents[av], oData);

                                        // console.log("aoLogicResolve");
                                        // console.log(aoLogicResolve);
    
                                        if (aoLogicResolve && aoLogicResolve.length) {

                                            //console.log(sAttributeName);

                                            let aoSubAttributes = F_LOOP_ATTRIBUTES(aoLogicResolve, sAttributeName, oData);
                                            
                                            // console.log("aoSubAttrobites");
                                            // console.log(aoSubAttributes);
                                           
                                        }
    
                                        break;
    
                                }
    
                            }
    
                        }
    
                        break;

                    case "logic":

                        let aoLogicResolve = this.resolveLogic(oNode, oData);
    
                        if (aoLogicResolve && aoLogicResolve.length) {
                            F_LOOP_ATTRIBUTES(aoLogicResolve, null, oData);
                        }

                        break;

                    case "data":
                    
                        console.log

                        break;

                    default:

                        if (sSpecificAttr) {

                            if (!oAttributes[sSpecificAttr]) {
                                oAttributes[sSpecificAttr] = [];
                            }

                            switch(oNode.type) {

                                case "literal":
            
                                    F_HANDLE_LITERAL_ATTR(oNode, sSpecificAttr);
                                    break;

                                case "data":

                                    F_HANDLE_DATA_ATTR(oNode, sSpecificAttr);
                                    break;
                            }

                        }
    
                }
                
            }

            if (Object.keys(oAttributes).length) {

                return oAttributes;
            }

            return null;
        }

        // Cal the intial attributes list
        F_LOOP_ATTRIBUTES(aoAttributes, null, oData);

        if (Object.keys(oAttributes).length) {

            return oAttributes;
        }

        return null;
    }

    generate(vTemplate = null, oLocalData) {

        let aoTemplate = null;

        if (vTemplate) {

            if (Array.isArray(vTemplate)) {

                aoTemplate = vTemplate;
            }
            else if (typeof vTemplate === "string") {

                if (this.mTemplates.has(vTemplate)) {

                    aoTemplate = this.mTemplates.get(vTemplate);
                }
                else {

                    console.error(`Unknown template was requested: ${vTemplate}`);
                }

            }
            else {

                console.error(`Unexpected data type sent in instead of template name or AST`);
            }

        }
        else {

            console.error(`You must provide a template AST or spcifiy a known template name`);
        }

        // Only continue, if you have a template loaded into memory
        if (aoTemplate) {

            let dRenderedNodes = document.createDocumentFragment();

            for (let a = 0, aLen = aoTemplate.length; a < aLen; a++) {

                let oNode = aoTemplate[a];

                switch (oNode.type) {

                    case "tag":
            
                        // Call the create element function
                        let dNewElement = document.createElement(oNode.name);

                        // Check and resolve attribute objects 
                        if (oNode.attributes && Array.isArray(oNode.attributes) && oNode.attributes.length) {
            
                            let oFinalizedAttributes = this.generateAttributes(oNode.attributes, oLocalData);
            
                            if (oFinalizedAttributes) {
            
                                for (let sAttributeName in oFinalizedAttributes) {
            
                                    switch(sAttributeName) {
            
                                        case "checked":
                                        case "multi":
                                        case "selected":
            
                                            console.warn(`valueless attributes not supported`);
                                            break;
            
                                        default:
            
                                            dNewElement.setAttribute(sAttributeName, oFinalizedAttributes[sAttributeName].join(" ").trim());
                                            break;
            
                                    }
            
                                }
            
                            }
                        }
            
                        // Check and render children
                        if (oNode.contents && Array.isArray(oNode.contents) && oNode.contents.length) {
            
                            let dChildren = this.generate(oNode.contents, oLocalData);

                            if (dChildren) {

                                // Loop through all children and add them to the parent (dNewEelement)
                                while (dChildren.hasChildNodes()) {  
                
                                    dNewElement.appendChild(dChildren.firstChild.cloneNode(true))
                
                                    dChildren.removeChild(dChildren.firstChild);
                                }

                            }
            
                        }
            
                        dRenderedNodes.appendChild(dNewElement);
                        break;

                    case "literal":

                        switch(typeof oNode.value) {

                            case "string":

                                dRenderedNodes.appendChild(document.createTextNode(`${oNode.value}`));
                                break;

                            case "object":

                                dRenderedNodes.appendChild(document.createTextNode(this.convertObjectToString(oNode.vlaue)));
                                break;

                            case "number":
                            case "boolean":

                                dRenderedNodes.appendChild(document.createTextNode(oNode.value.toString()));
                                break;
                        }

                        break;

                    case "data":

                        // In this location, the value is being called after or inside of a tag node

                        let sResolvedValue = this.resolveData(oNode, oLocalData);

                        switch (typeof sResolvedValue) {

                            case "object":

                                sResolvedValue = this.convertObjectToString(sResolvedValue);
                                break;

                            case "boolean":
                            case "number":

                                sResolvedValue = sResolvedValue.toString();
                                break;

                        }

                        dRenderedNodes.appendChild(document.createTextNode(sResolvedValue));

                        break;

                    case "logic":

                        // Catch logs
                        if (oNode.name === "log") {

                            console.log("Log logic was caught");

                        }
                        else {

                            // Resolve the logic node
                            let aoLogicResolve = this.resolveLogic(oNode, oLocalData);
                            
                            // If anything is returned, process it with a internal generate
                            if (aoLogicResolve && aoLogicResolve.length) {
    
                                let dChildren = this.generate(aoLogicResolve, oLocalData);
    
                                if (dChildren) {
    
                                    // Loop through all children and add them to the parent (dNewEelement)
                                    while (dChildren.hasChildNodes()) {  
                    
                                        dRenderedNodes.appendChild(dChildren.firstChild.cloneNode(true))
                    
                                        dChildren.removeChild(dChildren.firstChild);
                                    }
    
                                }
    
                            }
                            
                        }

                        break;

                    case "loop":

                        switch(oNode.name) {

                            case "for":

                                let vDataSet = this.resolveData(oNode.source, oLocalData);

                                console.log("For Loop");
                                console.log(vDataSet);

                                if (vDataSet !== null) {

                                    if (Array.isArray(vDataSet)) {

                                        let iIndex = 0;

                                        for (let vData of vDataSet) {

                                            let oLoopData = {
                                                "@index": iIndex,
                                                "@key": iIndex,
                                                "@source": vDataSet
                                            }

                                            if (typeof vData !== "object")  {
                                                oLoopData.value = vData; 
                                                oLoopData.this = vData;
                                            }

                                            let dLoopChildren = this.generate(oNode.contents, oLoopData);

                                            if (dLoopChildren.children) {

                                                while (dLoopChildren.hasChildNodes()) {  
                
                                                    dRenderedNodes.appendChild(dLoopChildren.firstChild.cloneNode(true))
                                
                                                    dLoopChildren.removeChild(dLoopChildren.firstChild);
                                                }
                                            }

                                        }

                                    }
                                    else if (typeof vDataSet === "object") {

                                        let iIndex = 0;

                                        for (let sObjectKey in vDataSet) {

                                            console.log(sObjectKey);

                                            let oLoopData = {
                                                "@index": iIndex,
                                                "@key": sObjectKey,
                                                "@source": vDataSet
                                            }

                                            if (typeof vDataSet[sObjectKey] !== "object")  {
                                                oLoopData.value = vData; 
                                                oLoopData.this = vData;
                                            }
                                            else {

                                                if (Array.isArray(vDataSet[sObjectKey])) {

                                                }
                                                else {

                                                    oLoopData = {
                                                        ...oLoopData,
                                                        ...vDataSet[sObjectKey]
                                                    }
                                                }

                                            }

                                            iIndex += 1;
                                        }

                                    }

                                }


                                break;

                        }

                        break;

                    case "template":

                        if (oNode.name && this.mTemplates.has(oNode.name)) {

                            let oNewDataObject = null;

                            if (oNode.arguments && oNode.arguments.length) {

                                for (let vArg of oNode.arguments) {

                                    switch (vArg.type) {

                                        case "data":

                                            if (vArg.value !== "this") {

                                                let oResolvedData = this.resolveData(vArg, oLocalData);

                                                if (oResolvedData !== null) {

                                                    if (oNewDataObject === null) {
                                                        oNewDataObject = {};
                                                    }

                                                    oNewDataObject = {
                                                        ...oResolvedData
                                                    };

                                                }

                                            }
                                            else {
                                             
                                                if (oNewDataObject === null) {
                                                    oNewDataObject = {};
                                                }

                                                oNewDataObject = {
                                                    ...oLocalData
                                                };
                                                
                                            }

                                            break;

                                        case "keyValue":

                                            if (vArg.contents && vArg.contents.length) {

                                                for (let oKeyValue of vArg.contents) {

                                                    switch(oKeyValue.type) {

                                                        case "literal":

                                                            if (oNewDataObject === null) {
                                                                oNewDataObject = {};
                                                            }

                                                            oNewDataObject[vArg.key] = oKeyValue.value;
                                                            break;

                                                        case "data":

                                                            if (oNewDataObject === null) {
                                                                oNewDataObject = {};
                                                            }

                                                            let oResolvedData = this.resolveData(oKeyValue, oLocalData);

                                                            if (oResolvedData) {
                                                                oNewDataObject[vArg.key] = oResolvedData;
                                                            }

                                                            break;

                                                    }


                                                }

                                            }

                                            break;

                                    }

                                }

                            }

                            let dTemplateResults = this.generate(this.mTemplates.get(oNode.name), oNewDataObject);
                            
                            if (dTemplateResults) {

                                // Loop through all children and add them to the parent (dNewEelement)
                                while (dTemplateResults.hasChildNodes()) {  
                
                                    dRenderedNodes.appendChild(dTemplateResults.firstChild.cloneNode(true))
                
                                    dTemplateResults.removeChild(dTemplateResults.firstChild);
                                }

                            }

                        }
                        else {

                            console.error(`Unknown template requested: ${oNode.name}`);
                        }

                        break;

                    default: 

                        console.log("Hit default");

                        break;
                }
            }

            return dRenderedNodes;
        }

        return null;

        // for (let a = 0, aLen = aoSource.length; a < aLen; a++) {

        //     let oNode = aoSource[a];

        //     switch (oNode.type) {
                            

        //             break;

        //         case "template":

        //             // verify this template is real first
        //             if (this.mTemplates.has(oNode.name)) {

        //                 // Create a new instance, passinf only the templates, and contexts we know about
        //                 let cSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

        //                 if (oNode.arguments && oNode.arguments.length) {

        //                     let oNewData = {};

        //                     for (let arg of oNode.arguments) {

        //                         if (typeof arg === "object") {

        //                             switch (arg.type) {

        //                                 case "data":

        //                                     let vData = this.resolveData(arg);

        //                                     if (typeof vData === "object" && !Array.isArray(vData)) {

        //                                         oNewData = {
        //                                             ...oNewData,
        //                                             ...vData
        //                                         };

        //                                     }
        //                                     else {

        //                                         console.log(`Template arguments can only be references to another object or a statically asigned key value pair where the value can be any data type.`)
        //                                     }

        //                                     break;

        //                                 case "keyValue":

        //                                     let sKey = arg.key;
        //                                     let vKeyValue = null;

        //                                     for (let oStaticContents of arg.contents) {

        //                                         switch(oStaticContents.type) {

        //                                             case "literal":
                                                    
        //                                                 if (vKeyValue) {
        //                                                     vKeyValue += oStaticContents.value;
        //                                                 }
        //                                                 else {

        //                                                     vKeyValue = oStaticContents.value;
        //                                                 }

        //                                                 break;

        //                                             case "data":

        //                                                 let vData = this.resolveData(oStaticContents);

        //                                                 if (vKeyValue) {

        //                                                     vKeyValue += vData;
        //                                                 }
        //                                                 else {

        //                                                     vKeyValue = vData;
        //                                                 }

        //                                                 break;

        //                                         }
                                                
        //                                     }

        //                                     oNewData[sKey] = vKeyValue;
        //                                     break;
        //                             }

        //                         }

        //                     }

        //                     let dSubChildren = cSubRender.generate(this.mTemplates.get(oNode.name), false, oNewData);

        //                     // Loop through all children and add them to the parent (dNewEelement)
        //                     while (dSubChildren.hasChildNodes()) {  

        //                         dParentNode.appendChild(dSubChildren.firstChild.cloneNode(true))

        //                         dSubChildren.removeChild(dSubChildren.firstChild);
        //                     }

                            
        //                     // Null out the template subrenderer
        //                     cSubRender = null;
        //                 }
        //                 else {

        //                     this.generate(this.mTemplates.get(oNode.name), dParentNode);
        //                 }
                        
        //             }
        //             else {

        //                 console.error(`Requested unknown template: ${oNode.name}`);
        //             }

        //             break;

        //         default:

        //             console.error(`Unknown tag was discovered: ${oNode.type}`);
        //             break;
        //     }

        // }

        // if (this.vTempData !== null) {
        //     this.vTempData = null;
        // }

        // return dParentNode;

    }
}